"""Vigenère: For demonstration use."""

from collections.abc import Sequence
from itertools import cycle
from typing import Any, Optional, TypeAlias, Union

# Letter is meant to be a single character or byte, but this is not enforced.
Letter: TypeAlias = Union[str, bytes]
Letters: TypeAlias = Sequence[Letter]


class Alphabet_meta(type):
    def __init__(cls, *args, **kwargs) -> None:  # type: ignore[no-untyped-def]
        cls._abc_caps_only = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"

        # Printable 7 bit ASCI with space but excluding backlash. Shuffled.
        cls._abc_shuffled_printable = r"""JDi-Km9247oBEctS%Isxz{<;=W^fL,[Y3Mgd6HV(kR8:_CF"*')>|#~Xay!]N+1vnqTl/}j$A.@0b ZGe`UPhp?Ow&ru5Q"""

        cls._default_alphabet = cls._abc_caps_only

    @property
    def default_alphabet(cls) -> Letters:
        return cls._default_alphabet

    @property
    def abc_caps_only(cls) -> Letters:
        return cls._abc_caps_only

    @property
    def abc_printable(cls) -> Letters:
        return cls._abc_shuffled_printable


class Alphabet(metaclass=Alphabet_meta):
    """An alphabet.

    This does not check if the alphabet is sensible. In particular, you may get
    very peculiar results if the alphabet contains duplicate elements.

    Instances of this class are conventionally immutable.
    """

    def __init__(
        self,
        alphabet: Optional[Letters] = None,
        prebaked: Optional[str] = None,
    ):
        """This does not check if the alphabet is sensible. In particular, you
        may get  very peculiar results if the alphabet contains duplicate
        elements.
        """

        match (alphabet, prebaked):
            case (None, None) | (None, "default"):
                abc = Alphabet.default_alphabet
            case (None, "caps"):
                abc = Alphabet.abc_caps_only
            case (None, "printable"):
                abc = Alphabet.abc_printable
            case (None, _):
                raise ValueError("Unknown pre-baked alphabet")
            case (_, None):
                if not isinstance(alphabet, Sequence):
                    raise TypeError("alphabet must be a Sequence")
                abc = alphabet
            case (_, _):
                raise ValueError(
                    "Can't use both explicit and pre-baked alphabet"
                )

        self._alphabet = abc

        self._modulus = len(self._alphabet)

        # Set up char to index table
        self._abc2idx: dict[Letter, int] = {
            c: i for i, c in enumerate(self._alphabet)
        }

    @property
    def alphabet(self) -> Letters:
        return self._alphabet

    @property
    def modulus(self) -> int:
        return self._modulus

    @property
    def abc2idx(self) -> dict[Letter, int]:
        return self._abc2idx

    # We will want to use 'in' for Alphabet instances
    def __contains__(self, item: Any) -> bool:
        return item in self.alphabet

    # annoyingly, the type str is also used for single character strings
    # add, inverse, subtract all deal with single characters
    def add(self, a: Letter, b: Letter) -> Letter:
        """Returns the modular sum of two characters."""
        if a not in self or b not in self:
            raise ValueError("argument not an element")
        idx = (self.abc2idx[a] + self.abc2idx[b]) % self.modulus
        return self.alphabet[idx]

    def inverse(self, c: Letter) -> Letter:
        """Returns the additive inverse of character c"""
        if c not in self:
            raise ValueError("argument not an element")
        idx = (self.modulus - self.abc2idx[c]) % self.modulus
        return self.alphabet[idx]

    def subtract(self, a: Letter, b: Letter) -> Letter:
        """Returns the character corresponding to a - b."""
        return self.add(a, self.inverse(b))


class Cipher:
    """A Vigenère Cipher is a key and an alphabet."""

    def __init__(
        self, key: Letters, alphabet: Alphabet | Letters | None = None
    ):
        if isinstance(alphabet, Alphabet):
            abc = alphabet
        else:
            abc = Alphabet(alphabet)

        self._alphabet = abc

        if not key:
            raise ValueError("key must not be empty")

        if any([k not in self._alphabet for k in key]):
            raise ValueError(
                "key must be comprised of characters in the alphabet"
            )
        self._key: Letters = key

    @property
    def alphabet(self) -> Alphabet:
        return self._alphabet

    @property
    def key(self) -> Letters:
        return self._key

    def crypt(self, text: Letters, mode: str) -> Letters:
        """{en,de}crypts text depending on mode"""

        match mode:
            case "encrypt":
                operation = self.alphabet.add
            case "decrypt":
                operation = self.alphabet.subtract
            case _:
                raise ValueError("mode must be 'encrypt' or 'decrypt")

        # TODO: Generalize this for streaming input and output
        output: list[Letter] = []

        for c, k in zip(text, cycle(self.key)):
            if c not in self.alphabet:
                result = c
            else:
                result = operation(c, k)

            output.append(result)

        if isinstance(text, bytes):
            return b"".join(output)
        return "".join(output)  # type: ignore[arg-type]

    def encrypt(self, plaintext: Letters) -> Letters:
        """Returns ciphertext."""

        return self.crypt(plaintext, mode="encrypt")

    def decrypt(self, ciphertext: Letters) -> Letters:
        """Returns plaintext."""

        return self.crypt(ciphertext, mode="decrypt")



import sys

import pytest
from toy_crypto import utils


class TestUtils:
    def test_bits(self) -> None:
        vectors = [
            (0b1101, [1, 0, 1, 1]),
            (1, [1]),
            (0, []),
            (0o644, [0, 0, 1, 0, 0, 1, 0, 1, 1]),
        ]
        for n, expected in vectors:
            bits = [bit for bit in utils.lsb_to_msb(n)]
            assert bits == expected

    def test_xor(self) -> None:
        vectors = [
            (b"dusk", b"dawn", bytes.fromhex("00 14 04 05")),
            (
                b"Attack at dawn!",
                bytes(10) + bytes.fromhex("00 14 04 05 00"),
                b"Attack at dusk!",
            ),
        ]

        for x, y, pad in vectors:
            r = utils.xor(x, y)
            assert r == pad


if __name__ == "__main__":
    sys.exit(pytest.main(args=[__file__]))



import sys

import pytest
from toy_crypto import vigenere

Letter = vigenere.Letter
Letters = vigenere.Letters


class TestCryptionDefault:
    """Encryption/decryption tests with default alphabet."""

    default_abc = vigenere.Alphabet(None)

    class TVector:
        def __init__(
            self, key: Letters, ptext: Letters, ctext: Letters
        ) -> None:
            self.key = key
            self.ptext = ptext
            self.ctext = ctext

    vectors = [
        TVector(
            "AAAA",
            "ONCE UPON A MIDNIGHT DREARY",
            "ONCE UPON A MIDNIGHT DREARY",
        ),
        TVector(
            "C",
            "ONCE UPON A MIDNIGHT DREARY",
            "QPEG WRQP C OKFPKIJV FTGCTA",
        ),
        TVector(
            "RAVEN",
            "ONCE UPON A MIDNIGHT DREARY",
            "FNXI LPJR R HMQEIBLG DMINIY",
        ),
    ]

    def test_encrypt(self) -> None:
        for tv in self.vectors:
            cipher = vigenere.Cipher(tv.key, self.default_abc)
            ctext = cipher.encrypt(tv.ptext)

            assert ctext == tv.ctext

    def test_decrypt(self) -> None:
        for tv in self.vectors:
            cipher = vigenere.Cipher(tv.key, self.default_abc)
            ptext = cipher.decrypt(tv.ctext)

            assert ptext == tv.ptext


class TestCryptionPrintable:
    """Encryption/decryption tests with default alphabet."""

    abc = vigenere.Alphabet(prebaked="printable")

    class TVector:
        def __init__(
            self, key: Letters, ptext: Letters, ctext: Letters
        ) -> None:
            self.key = key
            self.ptext = ptext
            self.ctext = ctext

    vectors = [
        TVector(
            [abc.alphabet[0]],
            "ONCE UPON A MIDNIGHT DREARY",
            "ONCE UPON A MIDNIGHT DREARY",
        ),
        TVector(
            "C",
            "ONCE UPON A MIDNIGHT DREARY",
            r"""VBryLMgVBL;LZ1FB1[UILF?y;?b""",
        ),
        TVector(
            "RAVEN",
            "ONCE UPON A MIDNIGHT DREARY",
            r"""gVP=:,+YA:xX$,+2&;'3=.G=VU7""",
        ),
    ]

    def test_encrypt(self) -> None:
        for tv in self.vectors:
            cipher = vigenere.Cipher(tv.key, self.abc)
            ctext = cipher.encrypt(tv.ptext)

            assert ctext == tv.ctext

    def test_decrypt(self) -> None:
        for tv in self.vectors:
            cipher = vigenere.Cipher(tv.key, self.abc)
            ptext = cipher.decrypt(tv.ctext)

            assert ptext == tv.ptext


if __name__ == "__main__":
    sys.exit(pytest.main(args=[__file__]))



from toy_crypto.nt import lcm, modinv


class PublicKey:
    def __init__(self, modulus: int, public_exponent: int) -> None:
        self._N = modulus
        self._e = public_exponent

    @property
    def N(self) -> int:
        return self._N

    @property
    def e(self) -> int:
        return self._e

    def encrypt(self, message: int) -> int:
        """raw encryption with neither padding nor nonce"""
        if message < 0:
            raise ValueError("Positive messages only")
        if not message < self._N:
            raise ValueError("Message too big")

        return pow(base=message, exp=self._e, mod=self._N)

    def __eq__(self, other: object) -> bool:
        """True when each has the same modulus and public exponent.

        When comparing to a PrivateKey, this compares only the public parts.
        """
        if isinstance(other, PublicKey):
            return self.e == other.e and self.N == other.N

        return NotImplemented


class PrivateKey:
    DEFAULT_E = 65537

    def __init__(self, p: int, q: int, pub_exponent: int = DEFAULT_E) -> None:
        self._p = p
        self._q = q
        self._e = pub_exponent

        self._N = self._p * self._q
        self._pubkey = PublicKey(self._N, self._e)

        self._dP = modinv(self._e, p - 1)
        self._dQ = modinv(self._e, (self._q - 1))
        self._qInv = modinv(self._q, self._p)

        self._d = self._compute_d()

    @property
    def pub_key(self) -> PublicKey:
        return self._pubkey

    @property
    def e(self) -> int:
        return self._e

    def __eq__(self, other: object) -> bool:
        """True iff keys are mathematically equivalent

        Private keys with internal differences can behave identically
        with respect to input and output. This comparison will return
        True when they are equivalent in this respect.

        When compared to a PublicKey, this compares only the public part.
        """
        if isinstance(other, PrivateKey):
            return self.pub_key == other.pub_key

        if isinstance(other, PublicKey):
            return self.pub_key == other

        return NotImplemented

    def _compute_d(self) -> int:
        λ = lcm(self._p - 1, self._q - 1)
        try:
            return modinv(self.e, λ)
        except ValueError:
            raise ValueError("Inverse of e mod λ does not exist")

    def decrypt(self, ciphertext: int) -> int:
        """Decrypt ciphertext"""

        if ciphertext < 1 or ciphertext >= self.pub_key.N:
            raise ValueError("ciphertext is out of range")

        # m =  pow(base=ciphertext, exp=self._d, mod=self._N)
        # but we will use the CRT
        # version comes from  rfc8017 §5.1.2

        m_1 = pow(ciphertext, self._dP, self._p)
        m_2 = pow(ciphertext, self._dQ, self._q)

        # I need to review CRT to see what this is for
        h = ((m_1 - m_2) * self._qInv) % self._p

        m = m_2 + self._q * h
        return m



from typing import Any, NewType, TypeGuard

Prob = NewType("Prob", float)


def is_prob(val: Any) -> TypeGuard[Prob]:
    """true if val is a float, s.t. 0.0 <= va <= 1.0"""
    if not isinstance(val, float):
        return False
    return val >= 0.0 and val <= 1.0


PositiveInt = NewType("PositiveInt", int)


def is_positive_int(val: Any) -> TypeGuard[PositiveInt]:
    """true if val is a float, s.t. 0.0 <= va <= 1.0"""
    if not isinstance(val, int):
        return False
    return val >= 1



"""Supplements Python secrets module"""

import math
import secrets
from collections.abc import MutableSequence
from typing import Any


def randrange(*args: int) -> int:
    """Same as random.randrange(), but uses RNG from secrets."""

    start = 0
    step = 1
    match len(args):
        case 1:
            stop = args[0]
        case 2:
            start = args[0]
            stop = args[1]
        case 3:
            start = args[0]
            stop = args[1]
            step = args[2]

        case _:
            raise TypeError("a more useful message should go here")

    diff = stop - start
    if diff < 1:
        raise ValueError("stop must be greater than start")

    if step < 1:
        raise ValueError("step must be positive")

    if diff == 1:
        return start

    if step >= diff:  # only the bottom of the range will be allowed
        return start

    r = secrets.randbelow(diff // step)
    r *= step
    r += start

    return r


def shuffle(x: MutableSequence[Any]) -> None:
    """Shuffles x in place using CSPRNG"""

    # Uses the "modern" Fisher-Yates shuffle from Knuth via
    # https://en.wikipedia.org/wiki/Fisher–Yates_shuffle#The_modern_algorithm

    n = len(x)
    if n < 2:
        return
    for i in range(n - 1):
        j = randrange(i, n)
        x[i], x[j] = x[j], x[i]


# from FullRandom example in
# https://docs.python.org/3/library/random.html#examples
def random() -> float:
    """returns a 32-bit float in [0.0, 1.0)"""

    mantissa = 0x10_0000_0000_0000 | secrets.randbits(52)
    exponent = -53
    x = 0
    while not x:
        x = secrets.randbits(32)
        exponent += x.bit_length() - 32
    return math.ldexp(mantissa, exponent)



"""Utility functions"""

import itertools
import math
from collections.abc import Iterator


def lsb_to_msb(n: int) -> Iterator[int]:
    """
    Iterator 0s and 1s representing bits of n, starting with the least significant bit.

    :raises TypeError: if n is not an integer.
    :raises ValueError: if n is negative.
    """
    if not isinstance(n, int):
        raise TypeError("n must be an integer")
    if n < 0:
        raise ValueError("n cannot be negative")

    while n > 0:
        yield n & 1
        n >>= 1


def digit_count(x: float, b: int = 10) -> int:
    """returns the number of digits (base b) in the integer part of x"""

    x = abs(x)
    result = math.floor(math.log(x, base=b) + 1)
    return result


def xor(m: bytes, pad: bytes) -> bytes:
    """Returns the xor of m with a (repeated) pad.

    The pad is repeated if it is shorter than m.
    This can be thought of as bytewise Vigenère
    """

    r: list[bytes] = [bytes([a ^ b]) for a, b in zip(m, itertools.cycle(pad))]

    return b"".join(r)
